/*
 * Copyright (c) 2021, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <raft/mr/device/allocator.hpp>
#include <raft/mr/device/buffer.hpp>

#include <cub/cub.cuh>

namespace raft {
namespace sparse {
namespace distance {

/**
 * Computes the maximum number of columns that can be stored
 * in shared memory in dense form with the given block size
 * and precision.
 * @return the maximum number of columns that can be stored in smem
 */
template <typename value_idx, typename value_t, int tpb = 1024>
inline int max_cols_per_block() {
  // max cols = (total smem available - cub reduction smem)
  return (raft::getSharedMemPerBlock() -
          ((tpb / raft::warp_size()) * sizeof(value_t))) /
         sizeof(value_t);
}

}  // namespace distance
}  // namespace sparse
}  // namespace raft
